#import "objc/runtime.h"
#import <objc/objc-runtime.h>

static id instanceOfRuntimeSuperClass;
static id instanceOfRuntimeSubClass;

static Class runtimeSuperClass;
static Class runtimeSubClass;

//skeleton of the dummy main method
static NSString *MainMethod(id self, SEL _cmd)
{
    NSString* className = NSStringFromClass([self class]);
    return [NSString stringWithFormat: @"This is a MAIN METHOD, from 'MainMethod' method from: %@", className];
}

//skeleton of the the dummy sub method
static NSString *SubMethod(id self, SEL _cmd)
{
    NSString* className = NSStringFromClass([self class]);
    return [NSString stringWithFormat: @"This is a SUB METHOD, from 'SubMethod' method from: %@", className];
}

//add the given method to the SuperClass by "easy" encoding, (without handwritten signature string)
void addMethodToSuperClassByEasyEncoding()
{
    //add the type by getting the info from another class
    Method description = class_getInstanceMethod([NSObject class], @selector(description));
    const char *types = method_getTypeEncoding(description);
    
    //now add the method (with signature and implementation)
    class_addMethod(runtimeSuperClass, @selector(description), (IMP)MainMethod, types);
}

//create a superclass and its instance
void createARuntimeSuperClass()
{
    //Create a superclass
    runtimeSuperClass = objc_allocateClassPair([NSObject class], "RuntimeSuperClass", 0);
    
    //Get classname of runtime class
    NSString* className = NSStringFromClass([runtimeSuperClass class]);
    
    addMethodToSuperClassByEasyEncoding();
    
    //register the class
    objc_registerClassPair(runtimeSuperClass);
    
    //create an instance of the RuntimeClass
    instanceOfRuntimeSuperClass = [[runtimeSuperClass alloc] init];
    
    NSLog(@"Instance of the SuperClass has been created...");
}

void overrideMethodInSubClassByHardEncoding(void)
{
    //add the type by getting the info from another class
    Method description = class_getInstanceMethod([NSObject class], @selector(description));
	
	//create the signature string for the method
    NSString *typesNS = [NSString stringWithFormat: @"%s%s%s%s",
                         @encode(NSUInteger),
                         @encode(id), @encode(SEL),
                         @encode(id)];
    const char *types = [typesNS UTF8String];
    
    //'override' the 'description' method with the other implementation
    class_addMethod(runtimeSubClass, @selector(description), (IMP)SubMethod, types);
}

void createARuntimeSubClass()
{
    //create subclass
    runtimeSubClass = objc_allocateClassPair([runtimeSuperClass class], "RuntimeSubClass", 0);
    
    //register the subclass
    objc_registerClassPair(runtimeSubClass);
    
    //create an instance of the subclass
    instanceOfRuntimeSubClass = [[runtimeSubClass alloc] init];
    
    //the subclass inherited the method from the superclass
    NSLog(@"1. call SubClass's method, it will be not yet overridden.");
    NSLog([instanceOfRuntimeSubClass description]);
    
    overrideMethodInSubClassByHardEncoding();
    
    NSLog(@"Instance of the SubClass has been created...");
}

void invokeSuperClassMethodFromSubClass()
{
    //create a superclass object from the subclass
    struct objc_super superClass = {
        .receiver = instanceOfRuntimeSubClass,
        .super_class = class_getSuperclass([instanceOfRuntimeSubClass class])
    };
    
    //call the superclass's method from subclass
    NSLog(@"2. call SuperClass's method, from SubClass.");
    NSLog(objc_msgSendSuper(&superClass, @selector(description)));
    
    //call the subclass's changed method
    NSLog(@"3. call SubClass's overridden method.");
    NSLog([instanceOfRuntimeSubClass description]);
}

void checkKindOfSuperClass()
{
    NSLog(@"\n");
    //check the kind of superclass's instance
    if ([instanceOfRuntimeSuperClass isKindOfClass:runtimeSuperClass])
        NSLog(@"The type of instanceOfRuntimeSuperclass is RuntimeSuperClass");
    else
        NSLog(@"The type if instanceOfRuntimeSuperClass is not RuntimeSuperClass");
}

void checkKindOfSubClass()
{
    NSLog(@"\n");
    //check the kind of subclass's instance
    if ([instanceOfRuntimeSubClass isKindOfClass:runtimeSubClass])
        NSLog(@"The type of instanceOfRuntimeSubclass is RuntimeSubClass");
    else
        NSLog(@"The type if instanceOfRuntimeSubClass is not RuntimeSubClass");
}

//check the kind of the given instances
void checkKindsOfInstances()
{
    checkKindOfSuperClass();
    checkKindOfSubClass();
}

void getMethodsOfClassInstances()
{
    NSLog(@"\n");
    unsigned int numberOfMethods;
    
    //get the method list of superclass
    Method* methods = class_copyMethodList([runtimeSuperClass class], &numberOfMethods);
    
    //print the methods from the list
    NSLog(@"The superclass has: %d method(s)", numberOfMethods);
    for (int i = 0; i < numberOfMethods; i++)
        NSLog(@"Method name: %@", NSStringFromSelector(method_getName(methods[i])));
    
    //get the method list from subclass
    methods = class_copyMethodList([runtimeSubClass class], &numberOfMethods);
    
    //print the methods
    NSLog(@"The subclass has: %d method(s)", numberOfMethods);
    for (int i = 0; i < numberOfMethods; i++)
        NSLog(@"Method name: %@", NSStringFromSelector(method_getName(methods[i])));
    
}

int main(int argc, const char * argv[])
{
    @autoreleasepool {
        createARuntimeSuperClass();
        createARuntimeSubClass();
       
        invokeSuperClassMethodFromSubClass();
        
        checkKindsOfInstances();
        getMethodsOfClassInstances();
    }
    return 0;
}